from .distn import RegressionDistn, ClassificationDistn
from .normal import Normal, NormalFixedVar
from .multivariate_normal import MultivariateNormal
from .lognormal import LogNormal
from .exponential import Exponential
from .categorical import k_categorical, Bernoulli



import numpy as np
from sklearn.utils import check_array

def Y_from_censored(T, E=None):
    if T is None:
        return None
    else:
        T = check_array(T,ensure_2d=False)
        T=T.reshape(T.shape[0])
    if T.dtype == [
        ("Event", "?"),
        ("Time", "<f8"),
    ]:  # already processed. Necessary for when d_score() calls score() as in LogNormalCRPScore
        return T
    if E is None:
        E = np.ones_like(T)
    else:
        E = check_array(E,ensure_2d=False)
        E = E.reshape(E.shape[0])
    Y = np.empty(dtype=[("Event", np.bool), ("Time", np.float64)], shape=T.shape[0])
    Y["Event"] = E.astype(np.bool)
    Y["Time"] = T.astype(np.float64)
    return Y



import numpy as np
import scipy as sp
import scipy.stats
from ngboost.distns import Normal
from ngboost.ngboost import NGBoost
from ngboost.scores import MLE, CRPS
from ngboost.learners import default_tree_learner, default_linear_learner
from ngboost.evaluation import *
from sklearn.metrics import r2_score
from matplotlib import pyplot as plt
from argparse import ArgumentParser


if __name__ == "__main__":

    argparser = ArgumentParser()
    argparser.add_argument("--lr", type=float, default=0.1)
    argparser.add_argument("--dataset", type=str, default="simulated")
    argparser.add_argument("--noise-lvl", type=float, default=0.25)
    argparser.add_argument("--distn", type=str, default="Normal")
    argparser.add_argument("--natural", action="store_true")
    argparser.add_argument("--score", type=str, default="CRPS")
    args = argparser.parse_args()

    np.random.seed(123)

    m, n = 1200, 50
    noise = np.random.randn(*(m, 1))
    beta1 = np.random.randn(n, 1)
    X = np.random.randn(m, n) / np.sqrt(n)
    Y = (X @ beta1 + args.noise_lvl * noise).squeeze()
    print(X.shape, Y.shape)

    X_train, X_test = X[:1000, :], X[1000:,]
    Y_train, Y_test = Y[:1000], Y[1000:]

    ngb = NGBoost(
        n_estimators=400,
        learning_rate=args.lr,
        Dist=Normal,
        Base=default_linear_learner,
        natural_gradient=args.natural,
        minibatch_frac=1.0,
        Score=eval(args.score)(),
        verbose=True,
        verbose_eval=100,
    )

    losses = ngb.fit(X_train, Y_train)
    forecast = ngb.pred_dist(X_test)
    print("R2:", r2_score(Y_test, forecast.loc))



import numpy as np
from ngboost.helpers import Y_from_censored


class Score:
    def total_score(self, Y, sample_weight=None):
        return np.average(self.score(Y), weights=sample_weight)

    def grad(self, Y, natural=True):
        grad = self.d_score(Y)
        if natural:
            metric = self.metric()
            grad = np.linalg.solve(metric, grad)
        return grad


class LogScore(Score):
    """
    Generic class for the log scoring rule.
    
    The log scoring rule is the same as negative log-likelihood: -log(P̂(y)), also known as the maximum likelihood estimator. This scoring rule has a default method for calculating the Riemannian metric.
    """

    def metric(self, n_mc_samples=100):
        grads = np.stack([self.d_score(Y) for Y in self.sample(n_mc_samples)])
        return np.mean(np.einsum("sik,sij->sijk", grads, grads), axis=0)

    # autofit method from d_score?


MLE = LogScore


class CRPScore(Score):
    """
    Generic class for the continuous ranked probability scoring rule.
    """

    pass


CRPS = CRPScore



from ngboost import NGBSurvival
from ngboost.distns import LogNormal
from sklearn.datasets import load_boston
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error
import numpy as np

if __name__ == "__main__":

    X, Y = load_boston(True)
    X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size=0.2)

    # introduce administrative censoring
    T_train = np.minimum(Y_train, 30)
    E_train = Y_train > 30

    ngb = NGBSurvival(Dist=LogNormal).fit(X_train, T_train, E_train)
    Y_preds = ngb.predict(X_test)
    Y_dists = ngb.pred_dist(X_test)

    # test Mean Squared Error
    test_MSE = mean_squared_error(Y_preds, Y_test)
    print("Test MSE", test_MSE)

    # test Negative Log Likelihood
    test_NLL = -Y_dists.logpdf(Y_test.flatten()).mean()
    print("Test NLL", test_NLL)



import numpy as np
from ngboost.distns import LogNormal, Exponential, Normal
from ngboost.api import NGBSurvival
from ngboost.scores import MLE, CRPS
from ngboost.learners import default_tree_learner, default_linear_learner
from ngboost.evaluation import (
    calibration_time_to_event,
    plot_calibration_curve,
    calibration_regression,
)
from sklearn.model_selection import train_test_split
from sklearn.metrics import r2_score
from matplotlib import pyplot as plt
from argparse import ArgumentParser


if __name__ == "__main__":

    argparser = ArgumentParser()
    argparser.add_argument("--n-estimators", type=int, default=100)
    argparser.add_argument("--lr", type=float, default=1.0)
    argparser.add_argument("--eps", type=float, default=1.0)
    args = argparser.parse_args()

    m, n = 1000, 5
    X = np.random.randn(m, n) / np.sqrt(n)
    Y = X @ np.ones((n,)) + 0.5 * np.random.randn(*(m,))
    T = X @ np.ones((n,)) + 0.5 * np.random.randn(*(m,)) + args.eps
    E = (T > Y).astype(int)

    print(X.shape, Y.shape, E.shape)
    print(f"Event rate: {np.mean(E):.2f}")

    X_tr, X_te, Y_tr, Y_te, T_tr, T_te, E_tr, E_te = train_test_split(
        X, Y, T, E, test_size=0.2
    )

    ngb = NGBSurvival(
        Dist=Exponential,
        n_estimators=args.n_estimators,
        learning_rate=args.lr,
        natural_gradient=True,
        Base=default_linear_learner,
        Score=MLE,
        verbose=True,
        verbose_eval=True,
    )
    train_losses = ngb.fit(X_tr, np.exp(np.minimum(Y_tr, T_tr)), E_tr)

    preds = ngb.pred_dist(X_te)
    print(f"R2: {r2_score(Y_te, np.log(preds.mean()))}")

    plt.hist(preds.mean(), range=(0, 10), bins=30, alpha=0.5, label="Pred")
    plt.hist(np.exp(Y_te), range=(0, 10), bins=30, alpha=0.5, label="True")
    plt.legend()
    plt.show()

    # since we simulated the data we fully observe all outcomes
    # calibration assuming complete observations
    pctles, observed, slope, intercept = calibration_regression(preds, Y_te)
    plot_calibration_curve(pctles, observed)
    plt.show()

    # calibration for partial observations
    pctles, observed, slope, intercept = calibration_time_to_event(
        preds, np.minimum(Y_te, T_te).squeeze(), E_te.squeeze()
    )
    plot_calibration_curve(pctles, observed)
    plt.show()



import scipy as sp
import numpy as np
from scipy.stats import expon as dist
from ngboost.distns import RegressionDistn
from ngboost.scores import LogScore, CRPScore

eps = 1e-10


class ExponentialLogScore(LogScore):
    def score(self, Y):
        E, T = Y["Event"], Y["Time"]
        cens = (1 - E) * np.log(1 - self.dist.cdf(T) + eps)
        uncens = E * self.dist.logpdf(T)
        return -(cens + uncens)

    def d_score(self, Y):
        E, T = Y["Event"], Y["Time"]
        cens = (1 - E) * T.squeeze() / self.scale
        uncens = E * (-1 + T.squeeze() / self.scale)
        return -(cens + uncens).reshape((-1, 1))

    def metric(self):
        FI = np.ones_like(self.scale)
        return FI[:, np.newaxis, np.newaxis]


class ExponentialCRPScore(CRPScore):
    def score(self, Y):
        E, T = Y["Event"], Y["Time"]
        score = T + self.scale * (2 * np.exp(-T / self.scale) - 1.5)
        score[E == 1] -= (
            0.5 * self.scale[E == 1] * np.exp(-2 * T[E == 1] / self.scale[E == 1])
        )
        return score

    def d_score(self, Y):
        E, T = Y["Event"], Y["Time"]
        deriv = 2 * np.exp(-T / self.scale) * (self.scale + T) - 1.5 * self.scale
        deriv[E == 1] -= np.exp(-2 * T[E == 1] / self.scale[E == 1]) * (
            0.5 * self.scale[E == 1] - T[E == 1]
        )
        return deriv.reshape((-1, 1))

    def metric(self):
        M = 0.5 * self.scale[:, np.newaxis, np.newaxis]
        return M


class Exponential(RegressionDistn):
    """
    Implements the exponential distribution for NGBoost.

    The exponential distribution has one parameters, scale. See scipy.stats.expon for details.
    This distribution has both LogScore and CRPScore implemented for it and both work with right-censored data
    """

    n_params = 1
    censored_scores = [ExponentialLogScore, ExponentialCRPScore]

    def __init__(self, params):
        self._params = params
        self.scale = np.exp(params[0])
        self.dist = dist(scale=self.scale)

    def __getattr__(self, name):
        if name in dir(self.dist):
            return getattr(self.dist, name)
        return None

    # should implement a `sample()` method

    @property
    def params(self):
        return {"scale": self.scale}

    def fit(Y):
        m, s = sp.stats.expon.fit(Y)
        return np.array([np.log(m + s)])



import scipy as sp
import numpy as np

from scipy.stats import norm as dist


eps = 1e-6


class MultivariateNormal(object):
    # TODO: make n_params general
    n_params = 5

    def __init__(self, params, temp_scale=1.0):
        self.n_params, self.N = params.shape
        self.p = int(0.5 * (np.sqrt(8 * self.n_params + 9) - 3))

        self.loc = params[: self.p, :].T

        self.L = np.zeros((self.p, self.p, self.N))
        self.L[np.tril_indices(self.p)] = params[self.p :, :]
        self.L = np.transpose(self.L, (2, 0, 1))
        self.cov = self.L @ np.transpose(self.L, (0, 2, 1))
        self.cov_inv = np.linalg.inv(self.cov)
        self.dCovdL = self.D_cov_D_L()

    def mean(self):
        return np.exp(self.loc[:, 0])

    def D_cov_D_L(self):
        # create commutation matrix
        commutation = np.zeros((self.p ** 2, self.p ** 2))
        ind = np.arange(self.p ** 2).reshape(self.p, self.p).T.flatten()
        commutation[np.arange(self.p ** 2), ind] = 1.0

        # compute Jacobian
        dCovdL = (np.identity(self.p ** 2) + commutation) @ np.kron(
            self.L, np.identity(self.p)
        )
        dCovdL = dCovdL.reshape(-1, 2, 2, 2, 2).swapaxes(-2, -1)
        return dCovdL

    def nll(self, Y):
        try:
            E = Y["Event"]
            T = np.log(Y["Time"] + eps)

            (
                mu0_given_1,
                mu1_given_0,
                var0_given_1,
                var1_given_0,
            ) = self.conditional_dist(T)

            cens = (1 - E) * (
                dist.logpdf(T, loc=self.loc[:, 1], scale=self.cov[:, 1, 1] ** 0.5)
                + np.log(
                    eps + 1 - dist.cdf(T, loc=mu0_given_1, scale=var0_given_1 ** 0.5)
                )
            )
            uncens = E * (
                dist.logpdf(T, loc=self.loc[:, 0], scale=self.cov[:, 0, 0] ** 0.5)
                + np.log(
                    eps + 1 - dist.cdf(T, loc=mu1_given_0, scale=var1_given_0 ** 0.5)
                )
            )
            return -(cens + uncens)
        except:
            diff = Y - self.loc
            M = diff[:, None, :] @ self.cov_inv @ diff[:, :, None]
            half_log_det = np.log(eps + np.diagonal(self.L, axis1=1, axis2=2)).sum(-1)
            const = self.p / 2 * np.log(eps + 2 * np.pi)
            logpdf = -const - half_log_det - 0.5 * M.flatten()
            return -logpdf

    def D_nll(self, Y_):
        try:
            E = Y_["Event"]
            T = np.log(Y_["Time"] + eps)

            (
                mu0_given_1,
                mu1_given_0,
                var0_given_1,
                var1_given_0,
            ) = self.conditional_dist(T)
            mu0 = self.loc[:, 0]
            mu1 = self.loc[:, 1]
            var0 = self.cov[:, 0, 0]
            var1 = self.cov[:, 1, 1]
            cov = self.cov[:, 0, 1]

            # reshape Jacobian
            tril_indices = np.tril_indices(self.p)
            J = self.dCovdL[:, :, :, tril_indices[0], tril_indices[1]]
            J = np.transpose(J, (0, 3, 1, 2)).reshape(self.N, -1, self.p ** 2)
            J = J.swapaxes(-2, -1)

            # compute grad mu
            D = np.zeros((self.N, self.n_params))
            diff0 = T - mu0
            diff1 = T - mu1
            Z0 = diff0 / (var0 ** 0.5 + eps)
            Z1 = diff1 / (var1 ** 0.5 + eps)
            Z0_1 = (T - mu0_given_1) / (var0_given_1 ** 0.5 + eps)
            Z1_0 = (T - mu1_given_0) / (var1_given_0 ** 0.5 + eps)
            pdf0 = dist.pdf(T, loc=mu0_given_1, scale=var0_given_1 ** 0.5)
            cdf0 = dist.cdf(T, loc=mu0_given_1, scale=var0_given_1 ** 0.5)
            pdf1 = dist.pdf(T, loc=mu1_given_0, scale=var1_given_0 ** 0.5)
            cdf1 = dist.cdf(T, loc=mu1_given_0, scale=var1_given_0 ** 0.5)
            cens_mu0 = (1 - E) * (pdf0 / (eps + 1 - cdf0))
            uncens_mu0 = E * (
                diff0 / (eps + var0) - pdf1 / (eps + 1 - cdf1) * (cov / (eps + var0))
            )
            uncens_mu1 = E * (pdf1 / (eps + 1 - cdf1))
            cens_mu1 = (1 - E) * (
                diff1 / (eps + var1) - pdf0 / (eps + 1 - cdf0) * (cov / (eps + var1))
            )
            D[:, 0] = -(cens_mu0 + uncens_mu0)
            D[:, 1] = -(cens_mu1 + uncens_mu1)

            # compute grad sigma
            D_sigma = np.zeros((self.N, self.p ** 2))

            cens_var0 = (1 - E) * (
                pdf0 / (eps + 1 - cdf0) * Z0_1 / (eps + 2 * var0_given_1 ** 0.5)
            )
            uncens_var0 = E * (
                0.5 * ((Z0 ** 2 - 1) / (eps + var0))
                + pdf1
                / (eps + 1 - cdf1)
                * (
                    -Z0 * (cov / (eps + var0 ** 1.5))
                    + 0.5
                    * Z1_0
                    / (eps + var1_given_0 ** 0.5)
                    * (cov / (eps + var0)) ** 2
                )
            )
            D_sigma[:, 0] = -(cens_var0 + uncens_var0)

            uncens_var1 = E * (
                pdf1 / (eps + 1 - cdf1) * Z1_0 / (eps + 2 * var1_given_0 ** 0.5)
            )
            cens_var1 = (1 - E) * (
                0.5 * ((Z1 ** 2 - 1) / (eps + var1))
                + pdf0
                / (eps + 1 - cdf0)
                * (
                    -Z1 * (cov / (eps + var1 ** 1.5))
                    + 0.5
                    * Z0_1
                    / (eps + var0_given_1 ** 0.5)
                    * (cov / (eps + var1)) ** 2
                )
            )
            D_sigma[:, 3] = -(cens_var1 + uncens_var1)

            uncens_cov = E * (
                pdf1
                / (eps + 1 - cdf1)
                * (
                    Z0 / (eps + var0 ** 0.5)
                    - Z1_0 / (eps + var1_given_0 ** 0.5) * (cov / (eps + var0))
                )
            )
            cens_cov = (1 - E) * (
                pdf0
                / (eps + 1 - cdf0)
                * (
                    Z1 / (eps + var1 ** 0.5)
                    - Z0_1 / (eps + var0_given_1 ** 0.5) * (cov / (eps + var1))
                )
            )
            D_sigma[:, 1] = -(cens_cov + uncens_cov) * 0.5
            D_sigma[:, 2] = -(cens_cov + uncens_cov) * 0.5

            D_L = J.swapaxes(-2, -1) @ D_sigma[:, :, None]
            D[:, self.p :] = D_L[..., 0]

            return D

        except:
            # reshape Jacobian
            tril_indices = np.tril_indices(self.p)
            J = self.dCovdL[:, :, :, tril_indices[0], tril_indices[1]]
            J = np.transpose(J, (0, 3, 1, 2)).reshape(self.N, -1, self.p ** 2)
            J = J.swapaxes(-2, -1)

            # compute grad mu
            D = np.zeros((self.N, J.shape[-1] + self.p))
            sigma_inv = np.linalg.inv(self.cov)
            diff = self.loc - Y_
            D[:, : self.p] = (sigma_inv @ diff[:, :, None])[..., 0]

            # compute grad sigma
            D_sigma = 0.5 * (
                sigma_inv
                - sigma_inv @ (diff[:, :, None] * diff[:, None, :]) @ sigma_inv
            )
            D_sigma = D_sigma.reshape(self.N, -1)
            D_L = J.swapaxes(-2, -1) @ D_sigma[:, :, None]
            D[:, self.p :] = D_L[..., 0]

            return D

    def fisher_info(self):
        # reshape Jacobian
        tril_indices = np.tril_indices(self.p)
        J = self.dCovdL[:, :, :, tril_indices[0], tril_indices[1]]

        FI = np.zeros((self.N, self.n_params, self.n_params))

        # compute FI mu
        FI[:, : self.p, : self.p] = self.cov_inv

        # compute FI sigma
        M = np.einsum("nij,njkl->nikl", self.cov_inv, J)
        M = np.einsum("nijx,njky->nikxy", M, M)
        FI[:, self.p :, self.p :] = 0.5 * np.trace(M, axis1=1, axis2=2)

        return FI

    def conditional_dist(self, Y):
        mu0 = self.loc[:, 0]
        mu1 = self.loc[:, 1]
        var0 = self.cov[:, 0, 0]
        var1 = self.cov[:, 1, 1]
        cov = self.cov[:, 0, 1]

        mu0_given_1 = mu0 + cov * (1 / (eps + var1)) * (Y - mu1)
        mu1_given_0 = mu1 + cov * (1 / (eps + var0)) * (Y - mu0)
        var0_given_1 = var0 - cov * (1 / (eps + var1)) * cov
        var1_given_0 = var1 - cov * (1 / (eps + var0)) * cov

        return mu0_given_1, mu1_given_0, var0_given_1, var1_given_0

    def fit(Y):
        try:
            E = Y["Event"]
            T = np.log(Y["Time"] + eps)
            # place holder
            m = np.array([8.0, 8.0])
            sigma = np.array([[2.0, 1.0], [1.0, 2.0]])
            L = sp.linalg.cholesky(sigma, lower=True)
            return np.concatenate([m, L[np.tril_indices(2)]])
        except:
            N, p = Y.shape
            m = Y.mean(axis=0)
            diff = Y - m
            sigma = 1 / N * (diff[:, :, None] @ diff[:, None, :]).sum(0)
            L = sp.linalg.cholesky(sigma, lower=True)
            return np.concatenate([m, L[np.tril_indices(p)]])



import numpy as np
import pandas as pd
from argparse import ArgumentParser
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import PolynomialFeatures
from dfply import *
from ngboost.distns import LogNormal, Exponential, MultivariateNormal
from ngboost.api import NGBSurvival
from ngboost.scores import MLE, CRPS
from ngboost.learners import default_tree_learner, default_linear_learner
from ngboost.evaluation import *
from sksurv.ensemble import GradientBoostingSurvivalAnalysis as GBSA
from sksurv.metrics import concordance_index_censored

np.random.seed(1)

base_name_to_learner = {
    "tree": default_tree_learner,
    "linear": default_linear_learner,
}


def Y_join(T, E):
    col_event = "Event"
    col_time = "Time"
    y = np.empty(dtype=[(col_event, np.bool), (col_time, np.float64)], shape=T.shape[0])
    y[col_event] = E.values
    y[col_time] = T.values
    return y


if __name__ == "__main__":

    argparser = ArgumentParser()
    argparser.add_argument("--dataset", type=str, default="flchain")
    argparser.add_argument("--distn", type=str, default="LogNormal")
    argparser.add_argument("--n-est", type=int, default=200)
    argparser.add_argument("--reps", type=int, default=1)
    argparser.add_argument("--lr", type=float, default=0.01)
    argparser.add_argument("--score", type=str, default="MLE")
    argparser.add_argument("--natural", action="store_true")
    argparser.add_argument("--base", type=str, default="tree")
    argparser.add_argument("--minibatch-frac", type=float, default=1.0)
    argparser.add_argument("--verbose", action="store_true")
    args = argparser.parse_args()

    # processing strategy from [chapfuwa et al 2019]
    # impute missing values with the most frequent
    # then one-hot encode categorical variables

    if args.dataset == "flchain":
        df = pd.read_csv("./data/surv/flchain.csv")
        E = df["death"]
        Y = df["futime"]
        X = (
            df
            >> drop(X.death, X.futime, X.chapter)
            >> mutate(mgus=X.mgus.astype(float), age=X.age.astype(float))
        )
        X = X[Y > 0]
        E = E[Y > 0]
        Y = Y[Y > 0]
        # Y = np.c_[np.log(T) - np.mean(np.log(T)), C]
        X_num = X.select_dtypes(include=["float"])
        X_cat = X.select_dtypes(exclude=["float"])
        imputer = SimpleImputer(strategy="median")
        X_num = imputer.fit_transform(X_num.values)
        imputer = SimpleImputer(strategy="most_frequent")
        X_cat = imputer.fit_transform(X_cat.values)
        encoder = OneHotEncoder(sparse=False)
        X_cat = encoder.fit_transform(X_cat)
        X = np.c_[X_num, X_cat]

    elif args.dataset == "support":
        df = pd.read_csv("./data/surv/support2.csv")
        df = df.rename(columns={"d.time": "dtime"})
        Y = df["dtime"]
        E = df["death"]
        # Y = np.c_[np.log(Y) - np.mean(np.log(Y)), C]
        df >>= drop(
            X.dtime,
            X.death,
            X.hospdead,
            X.prg2m,
            X.prg6m,
            X.dnr,
            X.dnrday,
            X.aps,
            X.sps,
            X.surv2m,
            X.surv6m,
            X.totmcst,
        )
        X_num = df.select_dtypes(include=["float", "int"])
        X_cat = df.select_dtypes(exclude=["float", "int"])
        imputer = SimpleImputer(strategy="median")
        X_num = imputer.fit_transform(X_num.values)
        imputer = SimpleImputer(strategy="most_frequent")
        X_cat = imputer.fit_transform(X_cat.values)
        encoder = OneHotEncoder(sparse=False)
        X_cat = encoder.fit_transform(X_cat)
        X = np.c_[X_num, X_cat]

    elif args.dataset == "sprint":
        df = pd.read_csv("data/surv/sprint-cut.csv")
        E = df["cvd"]
        Y = df["t_cvds"] / 365.25
        # Y = np.c_[np.log(Y) - np.mean(np.log(Y)), C]
        X = (df >> drop("cvd", "t_cvds", "INTENSIVE")).values

    print(
        "== Dataset=%s X.shape=%s Censorship=%.4f"
        % (args.dataset, str(X.shape), np.mean(1 - E))
    )

    for itr in range(args.reps):

        X_train, X_test, Y_train, Y_test, E_train, E_test = train_test_split(
            X, Y, E, test_size=0.2
        )
        X_train, X_val, Y_train, Y_val, E_train, E_val = train_test_split(
            X_train, Y_train, E_train, test_size=0.2
        )

        ngb = NGBSurvival(
            Dist=eval(args.distn),
            n_estimators=args.n_est,
            learning_rate=args.lr,
            natural_gradient=args.natural,
            verbose=args.verbose,
            minibatch_frac=1.0,
            Base=base_name_to_learner[args.base],
            Score=eval(args.score),
        )

        train_losses = ngb.fit(X_train, Y_train, E_train)
        forecast = ngb.pred_dist(X_test)
        train_forecast = ngb.pred_dist(X_train)
        print(
            "NGB score: %.4f (val), %.4f (train)"
            % (
                concordance_index_censored(
                    E_test.astype(bool), Y_test, -forecast.mean()
                )[0],
                concordance_index_censored(
                    E_train.astype(bool), Y_train, -train_forecast.mean()
                )[0],
            )
        )

        ##
        ## sksurv
        ##
        gbsa = GBSA(
            n_estimators=args.n_est,
            learning_rate=args.lr,
            subsample=args.minibatch_frac,
            verbose=args.verbose,
        )
        gbsa.fit(X_train, Y_join(Y_train, E_train))
        print(
            "GBSA score: %.4f (val), %.4f (train)"
            % (
                gbsa.score(X_test, Y_join(Y_test, E_test)),
                gbsa.score(X_train, Y_join(Y_train, E_train)),
            )
        )



import numpy as np
from ngboost import NGBClassifier
from ngboost.distns import Bernoulli
from ngboost.learners import default_tree_learner

from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score


if __name__ == "__main__":

    np.random.seed(12345)

    X, Y = load_breast_cancer(True)
    X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size=0.2)

    ngb = NGBClassifier(Dist=Bernoulli)
    ngb.fit(X_train, Y_train)

    preds = ngb.pred_dist(X_test)
    print("ROC:", roc_auc_score(Y_test, preds.probs[1]))



from warnings import warn

import numpy as np
from ngboost.helpers import Y_from_censored


class Distn(object):
    """
    User should define:
    - __init__(params) to hold self.params_ = params
    - X_scoring(self, Y) 
    - D_X_scoring(self, Y)
    - sample(self, n)
    - fit(Y)
    - predict(self) mean, mode, whatever (method to call for point prediction
    """

    def __init__(self, params):
        self._params = params

    def __getitem__(self, key):
        return self.__class__(self._params[:, key])

    def __len__(self):
        return self._params.shape[1]

    @classmethod
    def implementation(cls, Score, scores=None):
        """
        Finds the distribution-appropriate implementation of Score 
        (using the provided scores if cls.scores is empty)
        """
        if scores is None:
            scores = cls.scores
        if Score in scores:
            warn(
                f"Using Dist={Score.__name__} is unnecessary. NGBoost automatically selects the correct implementation when LogScore or CRPScore is used"
            )
            return Score
        else:
            try:
                return {S.__bases__[-1]: S for S in scores}[Score]
            except KeyError as err:
                raise ValueError(
                    f"The scoring rule {Score.__name__} is not implemented for the {cls.__name__} distribution."
                ) from err

    @classmethod
    def uncensor(cls, Score):
        DistScore = cls.implementation(Score, cls.censored_scores)

        class UncensoredScore(DistScore, DistScore.__base__):
            def score(self, Y):
                return super().score(Y_from_censored(Y))

            def d_score(self, Y):
                return super().d_score(Y_from_censored(Y))

        class DistWithUncensoredScore(cls):
            scores = [UncensoredScore]

        return DistWithUncensoredScore


class RegressionDistn(Distn):
    def predict(self):  # predictions for regression are typically conditional means
        return self.mean()


class ClassificationDistn(Distn):
    def predict(self):  # returns class assignments
        return np.argmax(self.class_probs(), 1)



from ngboost.distns import ClassificationDistn
from ngboost.scores import LogScore, CRPScore
import numpy as np
import scipy as sp
import scipy.special


class CategoricalLogScore(LogScore):
    def score(self, Y):
        return -np.log(self.probs[Y, range(len(Y))])

    def d_score(self, Y):
        return (self.probs.T - np.eye(self.K_)[Y])[:, 1 : self.K_]

    def metric(self):
        FI = -np.einsum(
            "ji,ki->ijk", self.probs[1 : self.K_, :], self.probs[1 : self.K_, :]
        )
        d = np.einsum("jii->ij", FI)
        d[:] += self.probs[1 : self.K_, :]
        return FI

    # a test:
    # if k==j:
    #     a= FI[i,j,k] == self.probs[k,i] - self.probs[k,i]*self.probs[j,i]
    # else:
    #     a= FI[i,j,k] == -self.probs[k,i]*self.probs[j,i]
    # a


class CategoricalCRPScore(CRPScore):
    def score(self, Y):
        return np.sum((self.probs - np.eye(self.K_)[Y]) ** 2, axis=1)

    def d_score(self, Y):
        return None

    def metric(self):
        return None


def k_categorical(K):
    """
    Factory function that generates classes for K-class categorical distributions for NGBoost

    The generated distribution has two parameters, loc and scale, which are the mean and standard deviation, respectively.
    This distribution has both LogScore and CRPScore implemented for it.
    """

    class Categorical(ClassificationDistn):

        scores = [CategoricalLogScore]
        problem_type = "classification"
        n_params = K - 1
        K_ = K

        def __init__(self, params):
            super().__init__(params)
            _, N = params.shape
            self.logits = np.zeros((K, N))
            self.logits[1:K, :] = params  # default the 0th class logits to 0
            self.probs = sp.special.softmax(self.logits, axis=0)
            # self.dist = dist(n=1, p=self.probs) # scipy doesn't allow vectorized multinomial (!?!?) why allow vectorized versions of the others?
            # this makes me want to refactor all the other code to use lists of distributions, would be more readable imo

        def fit(Y):
            _, n = np.unique(Y, return_counts=True)
            p = n / len(Y)
            return np.log(p[1:K]) - np.log(p[0])
            # https://math.stackexchange.com/questions/2786600/invert-the-softmax-function

        def sample1(self):  # this is just a helper for sample()
            cum_p = np.cumsum(self.probs, axis=0)[0:-1]
            interval = cum_p < np.random.random((1, len(self)))
            return np.sum(interval, axis=0)

        def sample(self, m):
            return np.array([self.sample1() for i in range(m)])

        def class_probs(self):  # required for any ClassificationDistn
            return self.probs.T

        @property
        def params(self):
            names = [f"p{j}" for j in range(self.n_params + 1)]
            return {name: p for name, p in zip(names, self.probs)}

    return Categorical


Bernoulli = k_categorical(2)



import scipy as sp
import numpy as np
from scipy.stats import lognorm as dist
from ngboost.distns import RegressionDistn
from ngboost.scores import LogScore, CRPScore


class LogNormalLogScoreCensored(LogScore):
    def score(self, Y):
        E = Y["Event"]
        T = Y["Time"]
        cens = (1 - E) * np.log(1 - self.dist.cdf(T) + self.eps)
        uncens = E * self.dist.logpdf(T)
        return -(cens + uncens)

    def d_score(self, Y):
        E = Y["Event"][:, np.newaxis]
        T = Y["Time"]
        lT = np.log(T)
        Z = (lT - self.loc) / self.scale

        D_uncens = np.zeros((self.loc.shape[0], 2))
        D_uncens[:, 0] = (self.loc - lT) / (self.scale ** 2)
        D_uncens[:, 1] = 1 - ((self.loc - lT) ** 2) / (self.scale ** 2)

        D_cens = np.zeros((self.loc.shape[0], 2))
        D_cens[:, 0] = -sp.stats.norm.pdf(lT, loc=self.loc, scale=self.scale) / (
            1 - self.dist.cdf(T) + self.eps
        )
        D_cens[:, 1] = (
            -Z
            * sp.stats.norm.pdf(lT, loc=self.loc, scale=self.scale)
            / (1 - self.dist.cdf(T) + self.eps)
        )

        return (1 - E) * D_cens + E * D_uncens

    def metric(self):
        FI = np.zeros((self.loc.shape[0], 2, 2))
        FI[:, 0, 0] = 1 / (self.scale ** 2) + self.eps
        FI[:, 1, 1] = 2
        return FI


class LogNormalCRPScoreCensored(CRPScore):
    def score(self, Y):
        E = Y["Event"]
        T = Y["Time"]
        lT = np.log(T)
        Z = (lT - self.loc) / self.scale

        crps_uncens = self.scale * (
            Z * (2 * sp.stats.norm.cdf(Z) - 1)
            + 2 * sp.stats.norm.pdf(Z)
            - 1 / np.sqrt(np.pi)
        )
        crps_cens = self.scale * (
            Z * sp.stats.norm.cdf(Z) ** 2
            + 2 * sp.stats.norm.cdf(Z) * sp.stats.norm.pdf(Z)
            - sp.stats.norm.cdf(np.sqrt(2) * Z) / np.sqrt(np.pi)
        )
        return (1 - E) * crps_cens + E * crps_uncens

    def d_score(self, Y):
        E = Y["Event"]
        T = Y["Time"]
        lT = np.log(T)
        Z = (lT - self.loc) / self.scale

        D = np.zeros((self.loc.shape[0], 2))
        D[:, 0] = E * -(2 * sp.stats.norm.cdf(Z) - 1) + (1 - E) * -(
            sp.stats.norm.cdf(Z) ** 2
            + 2 * Z * sp.stats.norm.cdf(Z) * sp.stats.norm.pdf(Z)
            + 2 * sp.stats.norm.pdf(Z) ** 2
            - 2 * sp.stats.norm.cdf(Z) * sp.stats.norm.pdf(Z) ** 2
            - np.sqrt(2 / np.pi) * sp.stats.norm.pdf(np.sqrt(2) * Z)
        )
        D[:, 1] = self.score(Y) + (lT - self.loc) * D[:, 0]
        return D

    def metric(self):
        I = np.zeros((self.loc.shape[0], 2, 2))
        I[:, 0, 0] = 2
        I[:, 1, 1] = self.scale ** 2
        I /= 2 * np.sqrt(np.pi)
        return I


class LogNormal(RegressionDistn):

    """
    Implements the log-normal distribution for NGBoost.

    The normal distribution has two parameters, s and scale (see scipy.stats.lognorm)
    This distribution has both LogScore and CRPScore implemented for it and both work for right-censored data.
    """

    n_params = 2
    censored_scores = [LogNormalLogScoreCensored, LogNormalCRPScoreCensored]

    def __init__(self, params):
        self._params = params
        self.loc = params[0]
        self.scale = np.exp(params[1])
        self.dist = dist(s=self.scale, scale=np.exp(self.loc))
        self.eps = 1e-5

    def __getattr__(self, name):
        if name in dir(self.dist):
            return getattr(self.dist, name)
        return None

    # should implmenent a `sample()` method

    @property
    def params(self):
        return {"s": self.scale, "scale": np.exp(self.loc)}

    def fit(Y):
        m, s = sp.stats.norm.fit(np.log(Y))
        return np.array([m, np.log(s)])



import pytest

from ngboost.distns import Normal, LogNormal, Exponential, Bernoulli, k_categorical
from ngboost.scores import LogScore, CRPScore
from sklearn.tree import DecisionTreeRegressor

from ngboost import NGBRegressor, NGBClassifier, NGBSurvival

from sklearn.datasets import load_boston, load_breast_cancer
from sklearn.model_selection import train_test_split

import numpy as np

# test all the dist methods and score implementation methods, i.e. they all return proper shapes and sizes and types
# check metric lines up with defaults for lognormal where applicable


@pytest.fixture(scope="module")
def learners():
    # add some learners that aren't trees
    return [
        DecisionTreeRegressor(criterion="friedman_mse", max_depth=5),
        DecisionTreeRegressor(criterion="friedman_mse", max_depth=3),
    ]


class TestRegDistns:
    @pytest.fixture(scope="class")
    def reg_dists(self):
        return {
            Normal: [LogScore, CRPScore],
            LogNormal: [LogScore, CRPScore],
            Exponential: [LogScore, CRPScore],
        }

    @pytest.fixture(scope="class")
    def reg_data(self):
        X, Y = load_boston(True)
        return train_test_split(X, Y, test_size=0.2)

    def test_dists(self, learners, reg_dists, reg_data):
        X_reg_train, X_reg_test, Y_reg_train, Y_reg_test = reg_data
        for Dist, Scores in reg_dists.items():
            for Score in Scores:
                for Learner in learners:
                    # test early stopping features
                    ngb = NGBRegressor(
                        Dist=Dist, Score=Score, Base=Learner, verbose=False
                    )
                    ngb.fit(X_reg_train, Y_reg_train)
                    y_pred = ngb.predict(X_reg_test)
                    y_dist = ngb.pred_dist(X_reg_test)
                    # test properties of output

    # test what happens when a dist that's not regression is passed in


class TestSurvDistns:
    @pytest.fixture(scope="class")
    def surv_dists(self):
        return {LogNormal: [LogScore, CRPScore], Exponential: [LogScore, CRPScore]}

    @pytest.fixture(scope="class")
    def surv_data(self):
        X, Y = load_boston(True)
        X_surv_train, X_surv_test, Y_surv_train, Y_surv_test = train_test_split(
            X, Y, test_size=0.2
        )

        # introduce administrative censoring to simulate survival data
        T_surv_train = np.minimum(Y_surv_train, 30)  # time of an event or censoring
        E_surv_train = (
            Y_surv_train > 30
        )  # 1 if T[i] is the time of an event, 0 if it's a time of censoring
        return X_surv_train, X_surv_test, T_surv_train, E_surv_train, Y_surv_test

    def test_dists(self, learners, surv_dists, surv_data):
        X_surv_train, X_surv_test, T_surv_train, E_surv_train, Y_surv_test = surv_data
        for Dist, Scores in surv_dists.items():
            for Score in Scores:
                for Learner in learners:
                    # test early stopping features
                    ngb = NGBSurvival(
                        Dist=Dist, Score=Score, Base=Learner, verbose=False
                    )
                    ngb.fit(X_surv_train, T_surv_train, E_surv_train)
                    y_pred = ngb.predict(X_surv_test)
                    y_dist = ngb.pred_dist(X_surv_test)
                    # test properties of output


class TestClsDistns:
    @pytest.fixture(scope="class")
    def cls_data(self):
        X, Y = load_breast_cancer(True)
        return train_test_split(X, Y, test_size=0.2)

    def test_bernoulli(self, learners, cls_data):
        X_cls_train, X_cls_test, Y_cls_train, Y_cls_test = cls_data
        for Learner in learners:
            # test early stopping features
            # test other args, n_trees, LR, minibatching- args as fixture
            ngb = NGBClassifier(
                Dist=Bernoulli, Score=LogScore, Base=Learner, verbose=False
            )
            ngb.fit(X_cls_train, Y_cls_train)
            y_pred = ngb.predict(X_cls_test)
            y_prob = ngb.predict_proba(X_cls_test)
            y_dist = ngb.pred_dist(X_cls_test)
            # test properties of output

    def test_categorical(self, learners, cls_data):
        X_cls_train, X_cls_test, Y_cls_train, Y_cls_test = cls_data
        for K in [2, 4, 7]:
            Dist = k_categorical(K)
            Y_cls_train = np.random.randint(0, K, (len(Y_cls_train)))

            for Learner in learners:
                # test early stopping features
                ngb = NGBClassifier(
                    Dist=Dist, Score=LogScore, Base=Learner, verbose=False
                )
                ngb.fit(X_cls_train, Y_cls_train)
                y_pred = ngb.predict(X_cls_test)
                y_prob = ngb.predict_proba(X_cls_test)
                y_dist = ngb.pred_dist(X_cls_test)
                # test properties of output


# test slicing and ._params



from warnings import warn


def manifold(Score, Distribution):
    """
    Mixes a scoring rule and a distribution together to create the resultant "Reimannian Manifold"
    (thus the name of the function). The resulting object has all the parameters of the distribution 
    can be sliced and indexed like one, and carries the distributions `fit` and `sample` methods, but 
    it also carries the appropriate `total_score` and `grad` methods that are inherited through 
    distribution-specific inheritence of the relevant implementation of the scoring rule
    """

    class Manifold(Distribution.implementation(Score), Distribution):
        pass

    return Manifold



from ngboost.distns import k_categorical
from ngboost import NGBClassifier
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import GridSearchCV, train_test_split
from sklearn.tree import DecisionTreeRegressor
from sklearn.linear_model import Ridge


if __name__ == "__main__":
    # An example where the base learner is also searched over (this is how you would vary tree depth):

    X, Y = load_breast_cancer(True)
    X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size=0.2)

    b1 = DecisionTreeRegressor(criterion="friedman_mse", max_depth=2)
    b2 = DecisionTreeRegressor(criterion="friedman_mse", max_depth=4)
    b3 = Ridge(alpha=0.0)

    param_grid = {
        "n_estimators": [20, 50],
        "minibatch_frac": [1.0, 0.5],
        "Base": [b1, b2],
    }

    ngb = NGBClassifier(natural_gradient=True, verbose=False, Dist=k_categorical(2))

    grid_search = GridSearchCV(ngb, param_grid=param_grid, cv=5)
    grid_search.fit(X_train, Y_train)
    print(grid_search.best_params_)



import numpy as np
from ngboost.distns import (
    Bernoulli,
    ClassificationDistn,
    LogNormal,
    Normal,
    RegressionDistn,
)
from ngboost.helpers import Y_from_censored
from ngboost.learners import default_tree_learner
from ngboost.ngboost import NGBoost
from ngboost.scores import LogScore
from sklearn.base import BaseEstimator
from sklearn.utils import check_array


class NGBRegressor(NGBoost, BaseEstimator):
    """
    Constructor for NGBoost regression models.

    NGBRegressor is a wrapper for the generic NGBoost class that facilitates regression. Use this class if you want to predict an outcome that could take an infinite number of (ordered) values.

    Parameters:
        Dist              : assumed distributional form of Y|X=x. A distribution from ngboost.distns, e.g. Normal
        Score             : rule to compare probabilistic predictions P̂ to the observed data y. A score from ngboost.scores, e.g. LogScore
        Base              : base learner to use in the boosting algorithm. Any instantiated sklearn regressor, e.g. DecisionTreeRegressor()
        natural_gradient  : logical flag indicating whether the natural gradient should be used
        n_estimators      : the number of boosting iterations to fit
        learning_rate     : the learning rate
        minibatch_frac    : the percent subsample of rows to use in each boosting iteration
        col_sample        : the percent subsample of columns to use in each boosting iteration
        verbose           : flag indicating whether output should be printed during fitting
        verbose_eval      : increment (in boosting iterations) at which output should be printed
        tol               : numerical tolerance to be used in optimization
        random_state      : seed for reproducibility. See https://stackoverflow.com/questions/28064634/random-state-pseudo-random-number-in-scikit-learn
    Output:
        An NGBRegressor object that can be fit.
    """

    def __init__(
        self,
        Dist=Normal,
        Score=LogScore,
        Base=default_tree_learner,
        natural_gradient=True,
        n_estimators=500,
        learning_rate=0.01,
        minibatch_frac=1.0,
        col_sample=1.0,
        verbose=True,
        verbose_eval=100,
        tol=1e-4,
        random_state=None,
    ):
        assert issubclass(
            Dist, RegressionDistn
        ), f"{Dist.__name__} is not useable for regression."

        if not hasattr(
            Dist, "scores"
        ):  # user is trying to use a dist that only has censored scores implemented
            Dist = Dist.uncensor(Score)

        super().__init__(
            Dist,
            Score,
            Base,
            natural_gradient,
            n_estimators,
            learning_rate,
            minibatch_frac,
            col_sample,
            verbose,
            verbose_eval,
            tol,
            random_state,
        )

    def __getstate__(self):
        state = super().__getstate__()
        # Remove the unpicklable entries.
        if self.Dist.__name__ == "DistWithUncensoredScore":
            state["Dist"] = self.Dist.__base__
            state["uncensor"] = True
        return state

    def __setstate__(self, state_dict):
        if "uncensor" in state_dict.keys():
            state_dict["Dist"] = state_dict["Dist"].uncensor(state_dict["Score"])
        super().__setstate__(state_dict)


class NGBClassifier(NGBoost, BaseEstimator):
    """
    Constructor for NGBoost classification models.

    NGBRegressor is a wrapper for the generic NGBoost class that facilitates classification. Use this class if you want to predict an outcome that could take a discrete number of (unordered) values.

    Parameters:
        Dist              : assumed distributional form of Y|X=x. A distribution from ngboost.distns, e.g. Bernoulli
        Score             : rule to compare probabilistic predictions P̂ to the observed data y. A score from ngboost.scores, e.g. LogScore
        Base              : base learner to use in the boosting algorithm. Any instantiated sklearn regressor, e.g. DecisionTreeRegressor()
        natural_gradient  : logical flag indicating whether the natural gradient should be used
        n_estimators      : the number of boosting iterations to fit
        learning_rate     : the learning rate
        minibatch_frac    : the percent subsample of rows to use in each boosting iteration
        col_sample        : the percent subsample of columns to use in each boosting iteration
        verbose           : flag indicating whether output should be printed during fitting
        verbose_eval      : increment (in boosting iterations) at which output should be printed
        tol               : numerical tolerance to be used in optimization
        random_state      : seed for reproducibility. See https://stackoverflow.com/questions/28064634/random-state-pseudo-random-number-in-scikit-learn
    Output:
        An NGBRegressor object that can be fit.
    """

    def __init__(
        self,
        Dist=Bernoulli,
        Score=LogScore,
        Base=default_tree_learner,
        natural_gradient=True,
        n_estimators=500,
        learning_rate=0.01,
        minibatch_frac=1.0,
        col_sample=1.0,
        verbose=True,
        verbose_eval=100,
        tol=1e-4,
        random_state=None,
    ):
        assert issubclass(
            Dist, ClassificationDistn
        ), f"{Dist.__name__} is not useable for classification."
        super().__init__(
            Dist,
            Score,
            Base,
            natural_gradient,
            n_estimators,
            learning_rate,
            minibatch_frac,
            col_sample,
            verbose,
            verbose_eval,
            tol,
            random_state,
        )

    def predict_proba(self, X, max_iter=None):
        """
        Probability prediction of Y at the points X=x

        Parameters:
            X        : numpy array of predictors (n x p)
            max_iter : get the prediction at the specified number of boosting iterations
            
        Output:
            Numpy array of the estimates of P(Y=k|X=x). Will have shape (n, K)
        """
        return self.pred_dist(X, max_iter=max_iter).class_probs()

    def staged_predict_proba(self, X, max_iter=None):
        """
        Probability prediction of Y at the points X=x at multiple boosting iterations

        Parameters:
            X        : numpy array of predictors (n x p)
            max_iter : largest number of boosting iterations to get the prediction for
            
        Output:
            A list of of the estimates of P(Y=k|X=x) of shape (n, K), one per boosting stage up to max_iter
        """
        return [
            dist.class_probs() for dist in self.staged_pred_dist(X, max_iter=max_iter)
        ]


class NGBSurvival(NGBoost, BaseEstimator):
    """
    Constructor for NGBoost survival models.

    NGBRegressor is a wrapper for the generic NGBoost class that facilitates survival analysis. Use this class if you want to predict an outcome that could take an infinite number of (ordered) values, but right-censoring is present in the observed data.

     Parameters:
        Dist              : assumed distributional form of Y|X=x. A distribution from ngboost.distns, e.g. LogNormal
        Score             : rule to compare probabilistic predictions P̂ to the observed data y. A score from ngboost.scores, e.g. LogScore
        Base              : base learner to use in the boosting algorithm. Any instantiated sklearn regressor, e.g. DecisionTreeRegressor()
        natural_gradient  : logical flag indicating whether the natural gradient should be used
        n_estimators      : the number of boosting iterations to fit
        learning_rate     : the learning rate
        minibatch_frac    : the percent subsample of rows to use in each boosting iteration
        col_sample        : the percent subsample of columns to use in each boosting iteration
        verbose           : flag indicating whether output should be printed during fitting
        verbose_eval      : increment (in boosting iterations) at which output should be printed
        tol               : numerical tolerance to be used in optimization
        random_state      : seed for reproducibility. See https://stackoverflow.com/questions/28064634/random-state-pseudo-random-number-in-scikit-learn
    Output:
        An NGBRegressor object that can be fit.
    """

    def __init__(
        self,
        Dist=LogNormal,
        Score=LogScore,
        Base=default_tree_learner,
        natural_gradient=True,
        n_estimators=500,
        learning_rate=0.01,
        minibatch_frac=1.0,
        col_sample=1.0,
        verbose=True,
        verbose_eval=100,
        tol=1e-4,
        random_state=None,
    ):

        assert issubclass(
            Dist, RegressionDistn
        ), f"{Dist.__name__} is not useable for regression."
        if not hasattr(Dist, "censored_scores"):
            raise ValueError(
                f"The {Dist.__name__} distribution does not have any censored scores implemented."
            )

        class SurvivalDistn(
            Dist
        ):  # Creates a new dist class from a given dist. The new class has its implemented scores
            scores = (
                Dist.censored_scores
            )  # set to the censored versions of the scores implemented for dist

            def fit(Y):  # and expects a {time, event} array as Y.
                return Dist.fit(Y["Time"])

        # assert issubclass(Dist, RegressionDistn), f'{Dist.__name__} is not useable for survival.'
        super().__init__(
            SurvivalDistn,
            Score,
            Base,
            natural_gradient,
            n_estimators,
            learning_rate,
            minibatch_frac,
            col_sample,
            verbose,
            verbose_eval,
            tol,
            random_state,
        )

    def fit(self, X, T, E, X_val=None, T_val=None, E_val=None, **kwargs):
        """
        Fits an NGBoost survival model to the data. For additional parameters see ngboost.NGboost.fit

        Parameters:
            X                       : DataFrame object or List or numpy array of predictors (n x p) in Numeric format
            T                       : DataFrame object or List or numpy array of times to event or censoring (n). Should be floats 
            E                       : DataFrame object or List or numpy array of event indicators (n). E[i] = 1 <=> T[i] is the time of an event, else censoring time
            T_val                   : DataFrame object or List or validation-set times, in Numeric format if any
            E_val                   : DataFrame object or List or validation-set event idicators, in Numeric format if any
        """
        
        X = check_array(X)
        
        if X_val is not None:
            X_val = check_array(X_val)
        
        return super().fit(
            X,
            Y_from_censored(T, E),
            X_val=X_val,
            Y_val=Y_from_censored(T_val, E_val),
            **kwargs,
        )



from scipy.stats import betabinom as dist
from scipy.stats import beta as betadist
import numpy as np
from ngboost.distns.distn import RegressionDistn
from ngboost.scores import LogScore
from scipy.special import polygamma, gamma, digamma
from scipy.special import beta as betafunction
#from fastbetabino import *
from array import array  
import sys   

class BetaBernoulliLogScore(LogScore): 
    
    def score(self, Y):
        return -self.dist.logpmf(Y)

    def d_score(self, Y):
        D = np.zeros((len(Y), 2)) # first col is dS/d(log(α)), second col is dS/d(log(β))

        D[:, 0] =   -self.alpha * (
                        digamma(self.alpha + self.beta) + 
                        digamma(Y + self.alpha) -
                        digamma(self.alpha + self.beta + 1) -
                        digamma(self.alpha)
                    )
        # D[:, 0] =   (
        #                 digamma(np.exp(self.m_raw)/(1 + np.exp(-self.mu_raw))) -
        #                 digamma(Y + np.exp(self.m_raw)/(1 + np.exp(-self.mu_raw))) -
        #                 digamma(np.exp(self.m_raw) - np.exp(self.m_raw)/(1 + np.exp(-self.mu_raw)))
        #                 digamma(-Y + np.exp(self.m_raw) + 1 - np.exp(self.m_raw)/(1 + np.exp(-self.mu_raw)))
        #             ) * np.exp(self.m_raw) * np.exp(-self.mu_raw) / (np.exp(self.mu_raw) + 1)**2
        D[:, 1] =   -self.beta * (
                        digamma(self.alpha + self.beta) + 
                        digamma(-Y + self.beta + 1) -
                        digamma(self.alpha + self.beta + 1) -
                        digamma(self.beta)
                    )
        # D[:, 1] =   (
        #                 np.exp(self.mu_raw) * (
        #                     digamma(np.exp(self.m_raw) / (1 + np.exp(-self.mu_raw))) -
        #                     digamma(Y + np.exp(self.m_raw) / (1 + np.exp(-self.mu_raw)))
        #                     digamma(np.exp(self.m_raw) + 1) -
        #                     digamma(np.exp(self.m_raw))
        #                 )
        #                 digamma(np.exp(self.m_raw) + 1)
        #                 digamma(np.exp(self.m_raw) - np.exp(self.m_raw) / (1 + np.exp(-self.mu_raw))) -
        #                 digamma(-Y + np.exp(self.m_raw) + 1 - np.exp(self.m_raw)/(1 + np.exp(-self.mu_raw))) -
        #                 digamma(np.exp(self.m_raw))
        #             ) * np.exp(self.m_raw) / (np.exp(self.mu_raw) + 1)
        return D

class BetaBernoulli(RegressionDistn):

    n_params = 2
    scores = [BetaBernoulliLogScore]

    def __init__(self, params):
        # save the parameters
        self._params = params
        
        # create other objects that will be useful later
        # self.mu_raw = params[0]
        # self.m_raw = params[1]
        # self.mu = self.sigmoid(self.mu_raw)
        # self.m = np.exp(self.m_raw)
        # self.alpha = self.mu * self.m # mu = alpha/(alpha + beta)
        # self.beta = self.m * (1 - self.mu) # m = alpha + beta
        # self.dist = dist(n=1, a=self.alpha, b=self.beta)
        self.log_alpha = params[0]
        self.log_beta = params[1]
        self.alpha = np.exp(self.log_alpha)
        self.beta = np.exp(self.log_beta)
        self.dist = dist(n=1, a=self.alpha, b=self.beta)

    def sigmoid(self, x):
        return 1/(1+np.exp(-x))

    def fit(Y):
        
        def fit_alpha_beta_py(impressions, clicks, alpha0=1.5, beta0=5, niter=1000):
            # based on https://github.com/lfiaschi/fastbetabino/blob/master/fastbetabino.pyx

            alpha_old=alpha0
            beta_old=beta0

            for it in range(niter):

                alpha=alpha_old*\
                (sum(digamma(c + alpha_old) - digamma(alpha_old) for c,i in zip(clicks,impressions)))/\
                (sum(digamma(i + alpha_old+beta_old) - digamma(alpha_old+beta_old) for c,i in zip(clicks,impressions)))


                beta=beta_old*\
                (sum(digamma(i-c + beta_old) - digamma(beta_old) for c,i in zip(clicks,impressions)))/\
                (sum(digamma(i + alpha_old+beta_old) - digamma(alpha_old+beta_old) for c,i in zip(clicks,impressions)))


                #print 'alpha {} | {}  beta {} | {}'.format(alpha,alpha_old,beta,beta_old)
                sys.stdout.flush()

                if np.abs(alpha-alpha_old) and np.abs(beta-beta_old)<1e-10:
                    #print 'early stop'
                    break

                alpha_old=alpha
                beta_old=beta

            return alpha, beta

        imps = np.ones_like(Y)
        alpha, beta = fit_alpha_beta_py(imps, Y) # use scipy's implementation
        return np.array([np.log(alpha), np.log(beta)])

    def sample(self, m):
        return np.array([self.dist.rvs() for i in range(m)])
    
    def __getattr__(self, name): # gives us access to Laplace.mean() required for RegressionDist.predict()
        if name in dir(self.dist):
            return getattr(self.dist, name)
        return None
    
    @property
    def params(self):
        return {'alpha':self.alpha, 'beta':self.beta}




from ngboost.distns import RegressionDistn
from ngboost.scores import LogScore, CRPScore
import scipy as sp
import numpy as np
from scipy.stats import norm as dist


class NormalLogScore(LogScore):
    def score(self, Y):
        return -self.dist.logpdf(Y)

    def d_score(self, Y):
        D = np.zeros((len(Y), 2))
        D[:, 0] = (self.loc - Y) / self.var
        D[:, 1] = 1 - ((self.loc - Y) ** 2) / self.var
        return D

    def metric(self):
        FI = np.zeros((self.var.shape[0], 2, 2))
        FI[:, 0, 0] = 1 / self.var
        FI[:, 1, 1] = 2
        return FI


class NormalCRPScore(CRPScore):
    def score(self, Y):
        Z = (Y - self.loc) / self.scale
        return self.scale * (
            Z * (2 * sp.stats.norm.cdf(Z) - 1)
            + 2 * sp.stats.norm.pdf(Z)
            - 1 / np.sqrt(np.pi)
        )

    def d_score(self, Y):
        Z = (Y - self.loc) / self.scale
        D = np.zeros((len(Y), 2))
        D[:, 0] = -(2 * sp.stats.norm.cdf(Z) - 1)
        D[:, 1] = self.score(Y) + (Y - self.loc) * D[:, 0]
        return D

    def metric(self):
        I = np.c_[
            2 * np.ones_like(self.var),
            np.zeros_like(self.var),
            np.zeros_like(self.var),
            self.var,
        ]
        I = I.reshape((self.var.shape[0], 2, 2))
        I = 1 / (2 * np.sqrt(np.pi)) * I
        return I


class Normal(RegressionDistn):
    """
    Implements the normal distribution for NGBoost.

    The normal distribution has two parameters, loc and scale, which are the mean and standard deviation, respectively.
    This distribution has both LogScore and CRPScore implemented for it.
    """

    n_params = 2
    scores = [NormalLogScore, NormalCRPScore]

    def __init__(self, params):
        super().__init__(params)
        self.loc = params[0]
        self.scale = np.exp(params[1])
        self.var = self.scale ** 2
        self.dist = dist(loc=self.loc, scale=self.scale)

    def fit(Y):
        m, s = sp.stats.norm.fit(Y)
        return np.array([m, np.log(s)])

    def sample(self, m):
        return np.array([self.rvs() for i in range(m)])

    def __getattr__(
        self, name
    ):  # gives us Normal.mean() required for RegressionDist.predict()
        if name in dir(self.dist):
            return getattr(self.dist, name)
        return None

    @property
    def params(self):
        return {"loc": self.loc, "scale": self.scale}


### Fixed Variance Normal ###
class NormalFixedVarLogScore(LogScore):
    def score(self, Y):
        return -self.dist.logpdf(Y)

    def d_score(self, Y):
        D = np.zeros((len(Y), 1))
        D[:, 0] = (self.loc - Y) / self.var
        return D

    def metric(self):
        FI = np.zeros((self.var.shape[0], 1, 1))
        FI[:, 0, 0] = 1 / self.var + 1e-5
        return FI


class NormalFixedVarCRPScore(CRPScore):
    def score(self, Y):
        Z = (Y - self.loc) / self.scale
        return self.scale * (
            Z * (2 * sp.stats.norm.cdf(Z) - 1)
            + 2 * sp.stats.norm.pdf(Z)
            - 1 / np.sqrt(np.pi)
        )

    def d_score(self, Y):
        Z = (Y - self.loc) / self.scale
        D = np.zeros((len(Y), 1))
        D[:, 0] = -(2 * sp.stats.norm.cdf(Z) - 1)
        return D

    def metric(self):
        I = np.c_[2 * np.ones_like(self.var)]
        I = I.reshape((self.var.shape[0], 1, 1))
        I = 1 / (2 * np.sqrt(np.pi)) * I
        return I


class NormalFixedVar(Normal):
    """
    Implements the normal distribution with variance=1 for NGBoost.

    The fixed-variance normal distribution has one parameters, loc which is the mean.
    This distribution has both LogScore and CRPScore implemented for it.
    """

    n_params = 1
    scores = [NormalFixedVarLogScore, NormalFixedVarCRPScore]

    def __init__(self, params):
        self.loc = params[0]
        self.var = np.ones_like(self.loc)
        self.scale = np.ones_like(self.loc)
        self.shape = self.loc.shape
        self.dist = dist(loc=self.loc, scale=self.scale)

    def fit(Y):
        m, s = sp.stats.norm.fit(Y)
        return m



import os
import setuptools


def get_version() -> str:
    version_filepath = os.path.join(os.path.dirname(__file__), "ngboost", "version.py")
    with open(version_filepath) as f:
        for line in f:
            if line.startswith("__version__"):
                return line.strip().split()[-1][1:-1]
    assert False


def get_long_description():
    readme_path = os.path.join(os.path.dirname(__file__), "README.md")
    with open(readme_path) as f:
        return f.read()


setuptools.setup(
    name="ngboost",
    version=get_version(),
    author="Stanford ML Group",
    author_email="avati@cs.stanford.edu",
    description="Library for probabilistic predictions via gradient boosting.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/stanfordmlgroup/ngboost",
    license="Apache License 2.0",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy>=1.17.2",
        "scipy>=1.3.1",
        "scikit-learn>=0.21.3",
        "tqdm>=4.36.1",
        "lifelines>=0.22.8",
    ],
    tests_require=["pytest", "pre-commit", "black"],
)



from sklearn.tree import DecisionTreeRegressor
from sklearn.linear_model import Ridge

default_tree_learner = DecisionTreeRegressor(
    criterion="friedman_mse",
    min_samples_split=2,
    min_samples_leaf=1,
    min_weight_fraction_leaf=0.0,
    max_depth=3,
    splitter="best",
    random_state=None,
)

default_linear_learner = Ridge(alpha=0.0, random_state=None)



import pytest

import numpy as np
from sklearn.model_selection import train_test_split

from ngboost import NGBClassifier, NGBRegressor
from ngboost.distns import Bernoulli, Normal


def test_classification():
    from sklearn.datasets import load_breast_cancer
    from sklearn.metrics import roc_auc_score, log_loss

    data, target = load_breast_cancer(True)
    x_train, x_test, y_train, y_test = train_test_split(
        data, target, test_size=0.2, random_state=42
    )
    ngb = NGBClassifier(Dist=Bernoulli, verbose=False)
    ngb.fit(x_train, y_train)
    preds = ngb.predict(x_test)
    score = roc_auc_score(y_test, preds)
    assert score >= 0.95

    preds = ngb.predict_proba(x_test)
    score = log_loss(y_test, preds)
    assert score <= 0.20

    score = ngb.score(x_test, y_test)
    assert score <= 0.20

    dist = ngb.pred_dist(x_test)
    assert isinstance(dist, Bernoulli)

    score = roc_auc_score(y_test, preds[:, 1])
    assert score >= 0.95


def test_regression():
    from sklearn.datasets import load_boston
    from sklearn.metrics import mean_squared_error

    data, target = load_boston(True)
    x_train, x_test, y_train, y_test = train_test_split(
        data, target, test_size=0.2, random_state=42
    )
    ngb = NGBRegressor(verbose=False)
    ngb.fit(x_train, y_train)
    preds = ngb.predict(x_test)
    score = mean_squared_error(y_test, preds)
    assert score <= 8.0

    score = ngb.score(x_test, y_test)
    assert score <= 8.0

    dist = ngb.pred_dist(x_test)
    assert isinstance(dist, Normal)

    score = mean_squared_error(y_test, preds)
    assert score <= 8.0



import numpy as np
from tqdm import tqdm
from lifelines import KaplanMeierFitter, NelsonAalenFitter
from matplotlib import pyplot as plt
from sklearn.metrics import roc_auc_score, r2_score, mean_squared_error


def calibration_regression(Forecast, Y, bins=11, eps=1e-3):
    """
    Calculate calibration in the regression setting.
    """
    pctles = np.linspace(eps, 1 - eps, bins)
    observed = np.zeros_like(pctles)
    for i, pctle in enumerate(pctles):
        icdfs = Forecast.ppf(pctle).reshape(Y.shape)
        observed[i] = np.mean(Y < icdfs)
    slope, intercept = np.polyfit(pctles, observed, deg=1)
    return pctles, observed, slope, intercept


def calibration_time_to_event(Forecast, T, E, bins=10, eps=1e-3):
    """
    Calculate calibration in the time-to-event setting, with integral transform and KM.
    """
    cdfs = Forecast.cdf(T)
    kmf = KaplanMeierFitter()
    kmf.fit(cdfs, E)
    idxs = np.round(np.linspace(0, len(kmf.survival_function_) - 1, 11))
    preds = np.array(kmf.survival_function_.iloc[idxs].index)
    obs = 1 - np.array(kmf.survival_function_.iloc[idxs].KM_estimate)
    slope, intercept = np.polyfit(preds, obs, deg=1)
    return preds, obs, slope, intercept


def calculate_calib_error(predicted, observed):
    return np.sum((predicted - observed) ** 2) / len(predicted)


def plot_pit_histogram(predicted, observed, **kwargs):
    plt.bar(
        x=predicted[1:],
        height=np.diff(observed),
        width=-np.diff(predicted),
        align="edge",
        fill=False,
        edgecolor="black",
        **kwargs
    )
    plt.xlim((0, 1))
    plt.xlabel("Probability Integral Transform")
    plt.ylabel("Density")
    plt.axhline(1.0 / (len(predicted) - 1), linestyle="--", color="grey")
    plt.title("PIT Histogram")


def plot_calibration_curve(predicted, observed):
    """
    Plot calibration curve.
    """
    slope, intercept = np.polyfit(predicted, observed, deg=1)
    plt.plot(predicted, observed, "o", color="black")
    plt.plot(
        np.linspace(0, 1),
        np.linspace(0, 1) * slope + intercept,
        "--",
        label="Slope: %.2f, Intercept: %.2f" % (slope, intercept),
        alpha=0.5,
        color="black",
    )
    plt.plot(np.linspace(0, 1), np.linspace(0, 1), "--", color="grey", alpha=0.5)
    plt.xlabel("Predicted CDF")
    plt.ylabel("Observed CDF")
    plt.title("Calibration Plot")
    plt.xlim((0, 1))
    plt.ylim((0, 1))
    plt.legend(loc="upper left")


def calculate_concordance_dead_only(preds, Y, E):
    """
    Calculate C-statistic for only cases where outcome is uncensored.
    """
    return calculate_concordance_naive(
        np.array(preds[E == 1]), np.array(Y[E == 1]), np.array(E[E == 1])
    )


def calculate_concordance_naive(preds, Y, E):
    """
    Calculate Harrell's C-statistic in the presence of censoring.

    Cases:
    - (c=0, c=0): both uncensored, can compare
    - (c=0, c=1): can compare if true censored time > true uncensored time
    - (c=1, c=0): can compare if true censored time > true uncensored time
    - (c=1, c=1): both censored, cannot compare
    """
    trues = Y
    concordance, N = 0, len(trues)
    counter = 0
    for i in tqdm(range(N)):
        for j in range(i + 1, N):
            if (
                (E[i] and E[j])
                or (E[i] and not E[j] and Y[i] < Y[j])
                or (not E[i] and E[j] and Y[i] > Y[j])
            ):
                if (preds[i] < preds[j] and trues[i] < trues[j]) or (
                    preds[i] > preds[j] and trues[i] > trues[j]
                ):
                    concordance += 1
                elif preds[i] == preds[j]:
                    concordance += 0.5
                counter += 1
    return concordance / counter



import numpy as np
import pandas as pd
from scipy.stats import norm as norm_dist
from argparse import ArgumentParser
from tqdm import tqdm
from sklearn.model_selection import train_test_split
from ngboost.distns import Normal, NormalFixedVar
from ngboost import NGBoost, NGBRegressor
from ngboost.scores import MLE, CRPS
from ngboost.learners import default_tree_learner, default_linear_learner
from sklearn.ensemble import GradientBoostingRegressor as GBR
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import KFold

np.random.seed(1)

dataset_name_to_loader = {
    "housing": lambda: pd.read_csv(
        "https://archive.ics.uci.edu/ml/machine-learning-databases/housing/housing.data",
        header=None,
        delim_whitespace=True,
    ),
    "concrete": lambda: pd.read_excel(
        "https://archive.ics.uci.edu/ml/machine-learning-databases/concrete/compressive/Concrete_Data.xls"
    ),
    "wine": lambda: pd.read_csv(
        "https://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/winequality-red.csv",
        delimiter=";",
    ),
    "kin8nm": lambda: pd.read_csv("data/uci/kin8nm.csv"),
    "naval": lambda: pd.read_csv(
        "data/uci/naval-propulsion.txt", delim_whitespace=True, header=None
    ).iloc[:, :-1],
    "power": lambda: pd.read_excel("data/uci/power-plant.xlsx"),
    "energy": lambda: pd.read_excel(
        "https://archive.ics.uci.edu/ml/machine-learning-databases/00242/ENB2012_data.xlsx"
    ).iloc[:, :-1],
    "protein": lambda: pd.read_csv("data/uci/protein.csv")[
        ["F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "RMSD"]
    ],
    "yacht": lambda: pd.read_csv(
        "http://archive.ics.uci.edu/ml/machine-learning-databases/00243/yacht_hydrodynamics.data",
        header=None,
        delim_whitespace=True,
    ),
    "msd": lambda: pd.read_csv("data/uci/YearPredictionMSD.txt").iloc[:, ::-1],
}

base_name_to_learner = {
    "tree": default_tree_learner,
    "linear": default_linear_learner,
}


if __name__ == "__main__":

    argparser = ArgumentParser()
    argparser.add_argument("--dataset", type=str, default="concrete")
    argparser.add_argument("--reps", type=int, default=5)
    argparser.add_argument("--n-est", type=int, default=2000)
    argparser.add_argument("--n-splits", type=int, default=20)
    argparser.add_argument("--distn", type=str, default="Normal")
    argparser.add_argument("--lr", type=float, default=0.01)
    argparser.add_argument("--natural", action="store_true")
    argparser.add_argument("--score", type=str, default="MLE")
    argparser.add_argument("--base", type=str, default="tree")
    argparser.add_argument("--minibatch-frac", type=float, default=None)
    argparser.add_argument("--verbose", action="store_true")
    args = argparser.parse_args()

    # load dataset -- use last column as label
    data = dataset_name_to_loader[args.dataset]()
    X, y = data.iloc[:, :-1].values, data.iloc[:, -1].values

    if not args.minibatch_frac:
        args.minibatch_frac = 1.0

    print(
        "== Dataset=%s X.shape=%s %s/%s"
        % (args.dataset, str(X.shape), args.score, args.distn)
    )

    y_true, ngb_rmse, ngb_nll = [], [], []

    if args.dataset == "msd":
        folds = [(np.arange(463715), np.arange(463715, len(X)))]
        args.minibatch_frac = 0.1
    else:
        kf = KFold(n_splits=args.n_splits)
        folds = kf.split(X)

        # Follow https://github.com/yaringal/DropoutUncertaintyExps/blob/master/UCI_Datasets/concrete/data/split_data_train_test.py
        n = X.shape[0]
        np.random.seed(1)
        folds = []
        for i in range(args.n_splits):
            permutation = np.random.choice(range(n), n, replace=False)
            end_train = round(n * 9.0 / 10)
            end_test = n

            train_index = permutation[0:end_train]
            test_index = permutation[end_train:n]
            folds.append((train_index, test_index))

    for itr, (train_index, test_index) in enumerate(folds):

        X_trainall, X_test = X[train_index], X[test_index]
        y_trainall, y_test = y[train_index], y[test_index]

        X_train, X_val, y_train, y_val = train_test_split(
            X_trainall, y_trainall, test_size=0.2
        )

        y_true += list(y_test.flatten())

        ngb = NGBRegressor(
            Base=base_name_to_learner[args.base],
            Dist=eval(args.distn),
            Score=eval(args.score),
            n_estimators=args.n_est,
            learning_rate=args.lr,
            natural_gradient=args.natural,
            minibatch_frac=args.minibatch_frac,
            verbose=args.verbose,
        )

        ngb.fit(X_train, y_train)

        # pick the best iteration on the validation set
        y_preds = ngb.staged_predict(X_val)
        y_forecasts = ngb.staged_pred_dist(X_val)

        val_rmse = [mean_squared_error(y_pred, y_val) for y_pred in y_preds]
        val_nll = [
            -y_forecast.logpdf(y_val.flatten()).mean() for y_forecast in y_forecasts
        ]
        best_itr = np.argmin(val_rmse) + 1

        # re-train using all the data after tuning number of iterations
        ngb = NGBRegressor(
            Base=base_name_to_learner[args.base],
            Dist=eval(args.distn),
            Score=eval(args.score),
            n_estimators=args.n_est,
            learning_rate=args.lr,
            natural_gradient=args.natural,
            minibatch_frac=args.minibatch_frac,
            verbose=args.verbose,
        )
        ngb.fit(X_trainall, y_trainall)

        # the final prediction for this fold
        forecast = ngb.pred_dist(X_test, max_iter=best_itr)
        forecast_val = ngb.pred_dist(X_val, max_iter=best_itr)

        # set the appropriate scale if using a homoskedastic Normal
        if args.distn == "NormalFixedVar":
            scale = (
                forecast.var * ((forecast_val.loc - y_val.flatten()) ** 2).mean() ** 0.5
            )
            forecast = norm_dist(loc=forecast.loc, scale=scale)

        ngb_rmse += [np.sqrt(mean_squared_error(forecast.mean(), y_test))]
        ngb_nll += [-forecast.logpdf(y_test.flatten()).mean()]

        print(
            "[%d/%d] BestIter=%d RMSE: Val=%.4f Test=%.4f NLL: Test=%.4f"
            % (
                itr + 1,
                args.n_splits,
                best_itr,
                np.sqrt(val_rmse[best_itr - 1]),
                np.sqrt(mean_squared_error(forecast.mean(), y_test)),
                ngb_nll[-1],
            )
        )

    print(
        "== RMSE GBM=%.4f +/- %.4f, NGB=%.4f +/- %.4f, NLL NGB=%.4f +/ %.4f"
        % (
            0.0,
            0.0,
            np.mean(ngb_rmse),
            np.std(ngb_rmse),
            np.mean(ngb_nll),
            np.std(ngb_nll),
        )
    )



from ngboost import NGBClassifier
from ngboost.distns import k_categorical
from ngboost.learners import default_tree_learner

from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import train_test_split

if __name__ == "__main__":

    X, y = load_breast_cancer(True)
    y[0:15] = 2  # artificially make this a 3-class problem instead of a 2-class problem
    X_train, X_test, Y_train, Y_test = train_test_split(X, y, test_size=0.2)

    ngb = NGBClassifier(
        Dist=k_categorical(3)
    )  # tell ngboost that there are 3 possible outcomes
    ngb.fit(X_train, Y_train)  # Y should have only 3 values: {0,1,2}

    # predicted probabilities of class 0, 1, and 2 (columns) for each observation (row)
    preds = ngb.predict_proba(X_test)



import numpy as np
import scipy as sp

from ngboost.scores import LogScore
from ngboost.distns import Normal, k_categorical
from ngboost.manifold import manifold
from ngboost.learners import default_tree_learner, default_linear_learner

from sklearn.utils import check_random_state,check_X_y,check_array
from sklearn.base import clone
from sklearn.tree import DecisionTreeRegressor

# import pdb


class NGBoost(object):
    """
    Constructor for all NGBoost models.

    This class implements the methods that are common to all NGBoost models. Unless you are implementing a new kind of regression (e.g. interval-censored, etc.), you should probably use one of NGBRegressor, NGBClassifier, or NGBSurvival.

    Parameters:
        Dist              : assumed distributional form of Y|X=x. A distribution from ngboost.distns, e.g. Normal
        Score             : rule to compare probabilistic predictions P̂ to the observed data y. A score from ngboost.scores, e.g. LogScore
        Base              : base learner to use in the boosting algorithm. Any instantiated sklearn regressor, e.g. DecisionTreeRegressor()
        natural_gradient  : logical flag indicating whether the natural gradient should be used
        n_estimators      : the number of boosting iterations to fit
        learning_rate     : the learning rate
        minibatch_frac    : the percent subsample of rows to use in each boosting iteration
        verbose           : flag indicating whether output should be printed during fitting
        verbose_eval      : increment (in boosting iterations) at which output should be printed
        tol               : numerical tolerance to be used in optimization
        random_state      : seed for reproducibility. See https://stackoverflow.com/questions/28064634/random-state-pseudo-random-number-in-scikit-learn
    Output:
        An NGBRegressor object that can be fit.
    """

    def __init__(
        self,
        Dist=Normal,
        Score=LogScore,
        Base=default_tree_learner,
        natural_gradient=True,
        n_estimators=500,
        learning_rate=0.01,
        minibatch_frac=1.0,
        col_sample=1.0,
        verbose=True,
        verbose_eval=100,
        tol=1e-4,
        random_state=None,
    ):
        self.Dist = Dist
        self.Score = Score
        self.Base = Base
        self.Manifold = manifold(Score, Dist)
        self.natural_gradient = natural_gradient
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.minibatch_frac = minibatch_frac
        self.col_sample = col_sample
        self.verbose = verbose
        self.verbose_eval = verbose_eval
        self.init_params = None
        self.base_models = []
        self.scalings = []
        self.col_idxs = []
        self.tol = tol
        self.random_state = check_random_state(random_state)
        self.best_val_loss_itr = None

    def __getstate__(self):
        state = self.__dict__.copy()
        # Remove the unpicklable entries.
        del state["Manifold"]
        if self.Dist.__name__ == "Categorical":
            del state["Dist"]
            state["K"] = self.Dist.n_params + 1
        return state

    def __setstate__(self, state_dict):
        if "K" in state_dict.keys():
            state_dict["Dist"] = k_categorical(state_dict["K"])
        state_dict["Manifold"] = manifold(state_dict["Score"], state_dict["Dist"])
        self.__dict__ = state_dict

    def fit_init_params_to_marginal(self, Y, sample_weight=None, iters=1000):
        self.init_params = self.Manifold.fit(
            Y
        )  # would be best to put sample weights here too
        return

    def pred_param(self, X, max_iter=None):
        m, n = X.shape
        params = np.ones((m, self.Manifold.n_params)) * self.init_params
        for i, (models, s, col_idx) in enumerate(
            zip(self.base_models, self.scalings, self.col_idxs)
        ):
            if max_iter and i == max_iter:
                break
            resids = np.array([model.predict(X[:, col_idx]) for model in models]).T
            params -= self.learning_rate * resids * s
        return params

    def sample(self, X, Y, sample_weight, params):
        idxs = np.arange(len(Y))
        col_idx = np.arange(X.shape[1])

        if self.minibatch_frac != 1.0:
            sample_size = int(self.minibatch_frac * len(Y))
            idxs = self.random_state.choice(
                np.arange(len(Y)), sample_size, replace=False
            )

        if self.col_sample != 1.0:
            col_size = int(self.col_sample * X.shape[1])
            col_idx = self.random_state.choice(
                np.arange(X.shape[1]), col_size, replace=False
            )

        weight_batch = None if sample_weight is None else sample_weight[idxs]

        return (
            idxs,
            col_idx,
            X[idxs, :][:, col_idx],
            Y[idxs],
            weight_batch,
            params[idxs, :],
        )

    def fit_base(self, X, grads, sample_weight=None):
        models = [
            clone(self.Base).fit(X, g, sample_weight=sample_weight) for g in grads.T
        ]
        fitted = np.array([m.predict(X) for m in models]).T
        self.base_models.append(models)
        return fitted

    def line_search(self, resids, start, Y, sample_weight=None, scale_init=1):
        S = self.Score
        D_init = self.Manifold(start.T)
        loss_init = D_init.total_score(Y, sample_weight)
        scale = scale_init

        # first scale up
        while True:
            scaled_resids = resids * scale
            D = self.Manifold((start - scaled_resids).T)
            loss = D.total_score(Y, sample_weight)
            norm = np.mean(np.linalg.norm(scaled_resids, axis=1))
            if not np.isfinite(loss) or loss > loss_init or scale > 256:
                break
            scale = scale * 2

        # then scale down
        while True:
            scaled_resids = resids * scale
            D = self.Manifold((start - scaled_resids).T)
            loss = D.total_score(Y, sample_weight)
            norm = np.mean(np.linalg.norm(scaled_resids, axis=1))
            if norm < self.tol:
                break
            if np.isfinite(loss) and loss < loss_init:
                break
            scale = scale * 0.5
        self.scalings.append(scale)
        return scale

    def fit(
        self,
        X,
        Y,
        X_val=None,
        Y_val=None,
        sample_weight=None,
        val_sample_weight=None,
        train_loss_monitor=None,
        val_loss_monitor=None,
        early_stopping_rounds=None,
    ):
        """
        Fits an NGBoost model to the data

        Parameters:
            X                       : DataFrame object or List or numpy array of predictors (n x p) in Numeric format
            Y                       : DataFrame object or List or numpy array of outcomes (n) in Numeric format. Should be floats for regression and integers from 0 to K-1 for K-class classification
            X_val                   : DataFrame object or List or numpy array of validation-set predictors in Numeric format, if any
            Y_val                   : DataFrame object or List or numpy array of validation-set outcomes in Numeric format, if any
            sample_weight           : how much to weigh each example in the training set. numpy array of size (n) (defaults to 1)
            val_sample_weight       : how much to weigh each example in the validation set. (defaults to 1)
            train_loss_monitor      : a custom score or set of scores to track on the training set during training. Defaults to the score defined in the NGBoost constructor
            val_loss_monitor        : a custom score or set of scores to track on the validation set during training. Defaults to the score defined in the NGBoost constructor
            early_stopping_rounds   : the number of consecutive boosting iterations during which the loss has to increase before the algorithm stops early

        Output:
            A fit NGBRegressor object
        """
        
        if Y is None:
            raise ValueError("y cannot be None")
        
        X, Y = check_X_y(X, Y, y_numeric=True)
        
        loss_list = []
        self.fit_init_params_to_marginal(Y)

        params = self.pred_param(X)
        if X_val is not None and Y_val is not None:
            X_val, Y_val = check_X_y(X_val, Y_val, y_numeric=True)
            val_params = self.pred_param(X_val)
            val_loss_list = []
            best_val_loss = np.inf

        if not train_loss_monitor:
            train_loss_monitor = lambda D, Y, W: D.total_score(Y, sample_weight=W)

        if not val_loss_monitor:
            val_loss_monitor = lambda D, Y: D.total_score(
                Y, sample_weight=val_sample_weight
            )

        for itr in range(self.n_estimators):
            _, col_idx, X_batch, Y_batch, weight_batch, P_batch = self.sample(
                X, Y, sample_weight, params
            )
            self.col_idxs.append(col_idx)

            D = self.Manifold(P_batch.T)

            loss_list += [train_loss_monitor(D, Y_batch, weight_batch)]
            loss = loss_list[-1]
            grads = D.grad(Y_batch, natural=self.natural_gradient)

            proj_grad = self.fit_base(X_batch, grads, weight_batch)
            scale = self.line_search(proj_grad, P_batch, Y_batch, weight_batch)

            # pdb.set_trace()
            params -= (
                self.learning_rate
                * scale
                * np.array([m.predict(X[:, col_idx]) for m in self.base_models[-1]]).T
            )

            val_loss = 0
            if X_val is not None and Y_val is not None:
                val_params -= (
                    self.learning_rate
                    * scale
                    * np.array(
                        [m.predict(X_val[:, col_idx]) for m in self.base_models[-1]]
                    ).T
                )
                val_loss = val_loss_monitor(self.Manifold(val_params.T), Y_val)
                val_loss_list += [val_loss]
                if val_loss < best_val_loss:
                    best_val_loss, self.best_val_loss_itr = val_loss, itr
                if (
                    early_stopping_rounds is not None
                    and len(val_loss_list) > early_stopping_rounds
                    and best_val_loss
                    < np.min(np.array(val_loss_list[-early_stopping_rounds:]))
                ):
                    if self.verbose:
                        print(f"== Early stopping achieved.")
                        print(
                            f"== Best iteration / VAL{self.best_val_loss_itr} (val_loss={best_val_loss:.4f})"
                        )
                    break

            if (
                self.verbose
                and int(self.verbose_eval) > 0
                and itr % int(self.verbose_eval) == 0
            ):
                grad_norm = np.linalg.norm(grads, axis=1).mean() * scale
                print(
                    f"[iter {itr}] loss={loss:.4f} val_loss={val_loss:.4f} scale={scale:.4f} "
                    f"norm={grad_norm:.4f}"
                )

            if np.linalg.norm(proj_grad, axis=1).mean() < self.tol:
                if self.verbose:
                    print(f"== Quitting at iteration / GRAD {itr}")
                break

        self.evals_result = {}
        metric = self.Score.__name__.upper()
        self.evals_result["train"] = {metric: loss_list}
        if X_val is not None and Y_val is not None:
            self.evals_result["val"] = {metric: val_loss_list}

        return self

    def score(self, X, Y):  # for sklearn
        return self.Manifold(self.pred_dist(X)._params).total_score(Y)

    def pred_dist(self, X, max_iter=None):
        """
        Predict the conditional distribution of Y at the points X=x

        Parameters:
            X        : DataFrame object or List or numpy array of predictors (n x p) in Numeric Format
            max_iter : get the prediction at the specified number of boosting iterations

        Output:
            A NGBoost distribution object
        """
        
        X = check_array(X)
        
        if (
            max_iter is not None
        ):  # get prediction at a particular iteration if asked for
            dist = self.staged_pred_dist(X, max_iter=max_iter)[-1]
        else:
            params = np.asarray(self.pred_param(X, max_iter))
            dist = self.Dist(params.T)
        return dist

    def staged_pred_dist(self, X, max_iter=None):
        """
        Predict the conditional distribution of Y at the points X=x at multiple boosting iterations

        Parameters:
            X        : numpy array of predictors (n x p)
            max_iter : largest number of boosting iterations to get the prediction for

        Output:
            A list of NGBoost distribution objects, one per boosting stage up to max_iter
        """
        predictions = []
        m, n = X.shape
        params = np.ones((m, self.Dist.n_params)) * self.init_params
        for i, (models, s, col_idx) in enumerate(
            zip(self.base_models, self.scalings, self.col_idxs)
        ):
            resids = np.array([model.predict(X[:, col_idx]) for model in models]).T
            params -= self.learning_rate * resids * s
            dists = self.Dist(
                np.copy(params.T)
            )  # if the params aren't copied, param changes with stages carry over to dists
            predictions.append(dists)
            if max_iter and i == max_iter:
                break
        return predictions

    def predict(self, X, max_iter=None):
        """
        Point prediction of Y at the points X=x

        Parameters:
            X        : DataFrame object or List or numpy array of predictors (n x p) in Numeric Format
            max_iter : get the prediction at the specified number of boosting iterations

        Output:
            Numpy array of the estimates of Y
        """
        
        X = check_array(X)
        
        return self.pred_dist(X, max_iter=max_iter).predict()

    def staged_predict(self, X, max_iter=None):
        """
        Point prediction of Y at the points X=x at multiple boosting iterations

        Parameters:
            X        : numpy array of predictors (n x p)
            max_iter : largest number of boosting iterations to get the prediction for

        Output:
            A list of numpy arrays of the estimates of Y, one per boosting stage up to max_iter
        """
        return [dist.predict() for dist in self.staged_pred_dist(X, max_iter=max_iter)]

    @property
    def feature_importances_(self):
        """
        Return the feature importances for all parameters in the distribution
            (the higher, the more important the feature).

        Returns
        -------
        feature_importances_ : array, shape = [n_params, n_features]
            The summation along second axis of this array is an array of ones,
            unless all trees are single node trees consisting of only the root
            node, in which case it will be an array of zeros.
        """
        # Check whether the model is fitted
        if not self.base_models:
            return None
        # Check whether the base model is DecisionTreeRegressor
        if not isinstance(self.base_models[0][0], DecisionTreeRegressor):
            return None
        # Reshape the base_models
        params_trees = zip(*self.base_models)
        # Get the feature_importances_ for all the params and all the trees
        all_params_importances = [
            [
                getattr(tree, "feature_importances_")
                for tree in trees
            ]
            for trees in params_trees
        ]

        if not all_params_importances:
            return np.zeros(
                len(self.base_models[0]),
                self.base_models[0][0].n_features_,
                dtype=np.float64,
            )
        # Weighted average of importance by tree scaling factors
        all_params_importances = np.average(
            all_params_importances, axis=1, weights=self.scalings
        )
        return all_params_importances / np.sum(
            all_params_importances, axis=1, keepdims=True
        )
